import pandas as pd
import numpy as np
from pandas import DataFrame
from typing import List, Tuple
from sklearn.base import RegressorMixin

from .data import Dataset
from .nda_typing import MatrixFloat64


def cal_padj(group: DataFrame) -> DataFrame:
    """
    Calculate the padj for a given group of p-values
    """
    n = len(group)
    group = group.sort_values("pval", ascending=True)
    group["padj"] = group["pval"] * n / np.arange(1, n + 1)
    return group


def skip_padj(group: DataFrame) -> DataFrame:
    """
    Skip the padj calculation for a given group of p-values
    """
    group["padj"] = group["pval"]
    return group


def proc_train_res(
    train_res: List[List[MatrixFloat64 | None]],
    models: List[RegressorMixin],
    dataset: Dataset,
    padj: bool = False,
) -> DataFrame:
    """
    Integrate the results from different chunks and calculate padj

    Parameters
    ----------
    result : List[List[MatrixFloat64 | None]]
        The result from the regression models, each Matrix element is a list of (pcc, pval)
    models : List[RegressorMixin]
        The list of regression models
    dataset : Dataset
        The dataset containing the information

    Returns
    -------
    DataFrame
        The integrated DataFrame containing the correlation, p-value, padj, and gene information
    """
    met_matrix, gene_idx = [], []
    for chunk in train_res:
        for result in chunk:
            if result is not None:
                gene_idx.append(True)
                met_matrix.append(result)
            else:
                gene_idx.append(False)
    if met_matrix is None or len(met_matrix) == 0:
        raise ValueError("No valid results found in the training results.")
    res = DataFrame(np.array(met_matrix).reshape(-1, 2))
    res.columns = ["corr", "pval"]
    model_names = [model.__name__ for model in models]
    res["model"] = res.index.map(lambda idx: model_names[idx % len(model_names)])
    not_na_genes = dataset.gene.name[gene_idx]
    res["gene"] = res.index.map(lambda idx: not_na_genes[idx // len(model_names)])
    res = res.astype(
        {
            "corr": np.float64,
            "pval": np.float64,
            "gene": np.str_,
            "model": "category",
        }
    )

    if padj:
        padj_func = cal_padj
    else:
        padj_func = skip_padj

    res = (
        res.groupby("model", observed=False)
        .apply(padj_func, include_groups=False)
        .reset_index(level="model")
        .dropna()
        .groupby("gene", observed=False)
        .apply(
            lambda group: group.loc[group["padj"].idxmin()],
            include_groups=False,
        )
        .reset_index(level="gene")
        .drop(columns=["pval"])
        .reset_index(drop=True)
        .merge(
            dataset.gene.df.groupby("gene")
            .apply(
                lambda group: group.assign(
                    start_min=group["start"].min(),
                    end_max=group["end"].max(),
                    center=(group["start"].min() + group["end"].max()) / 2,
                ),
                include_groups=False,
            )
            .filter(["chr", "start_min", "end_max", "center"])
            .reset_index(level="gene")
            .drop_duplicates(),
            how="left",
            left_on="gene",
            right_on="gene",
        )
        .astype({"chr": "string"})
        .assign(padj_norm=lambda df: -np.log10(df["padj"]))
    )

    res["padj_norm"] = -np.log10(res["padj"])
    finite_mask = np.isfinite(res["padj_norm"].to_numpy())
    if finite_mask.any():
        finite_max = float(res.loc[finite_mask, "padj_norm"].max())
        capped = res["padj_norm"].to_numpy()
        capped[~finite_mask] = finite_max + 1.0
        res["padj_norm"] = capped

    return res


def cal_sliding_window(
    met: DataFrame, chrom: str, window_size: int, step: int
) -> MatrixFloat64:
    """
    Sliding window to calculate the mean of the padj_norm values

    Parameters
    ----------
    met : DataFrame
        The DataFrame containing the padj_norm values
    chrom : str
        The chromosome to calculate the mean
    window_size : int
        The size of the window
    step : int
        The step size for the sliding window

    Returns
    -------
    MatrixFloat64
        The mean of the padj_norm values for each window and the start and end positions
    """
    met_chr = (
        met[met["chr"] == chrom]
        .sort_values(by=["start_min"], ascending=True)
        .reset_index()
    )
    window_mean = []
    gene_num = len(met_chr)
    for i in range(0, gene_num, step):
        start, end = i, i + window_size - 1
        end = end if end < gene_num else gene_num - 1
        if window_mean and window_mean[-1][1] == end:
            break
        window_mean.append(
            np.array([start, end, met_chr.loc[start:end, "padj_norm"].mean()])
        )
    return np.array(window_mean)


def merge_window(
    window_mean: MatrixFloat64, threshold: np.float64
) -> MatrixFloat64 | None:
    """
    Merge genes in the same region

    Parameters
    ----------
    window_mean : MatrixFloat64
        The mean of the padj_norm values for each window and the start and end positions
    threshold : np.float64
        The threshold to filter the mean values

    Returns
    -------
    MatrixFloat64
        The merged windows with start and end positions and the mean padj_norm value
    """
    window_loc = window_mean[window_mean[:, 2] >= threshold][:, 0:2]
    if len(window_loc) == 0:
        return None
    loc_merged, tmp = [], window_loc[0]
    for start, end in window_loc[1:]:
        if start <= tmp[1]:
            tmp = np.array([tmp[0], end])
        else:
            loc_merged.append(tmp)
            tmp = np.array([start, end])
    loc_merged.append(tmp)
    return np.array(loc_merged)


def significance(
    sliding_window_result: List[Tuple[str, MatrixFloat64, MatrixFloat64]],
    result: DataFrame,
) -> DataFrame:
    """
    Get the gene in the peek window of the green region

    Parameters
    ----------
    sliding_window_result : List[Tuple[str, MatrixFloat64, MatrixFloat64]]
        Results of the sliding window calculation
    result : DataFrame
        Integrated training results

    Returns
    -------
    DataFrame
        Gene table of the green region in the graph
    """
    region_gene = pd.DataFrame()
    for chr, window_mean, window_merged in sliding_window_result:
        if window_merged is None or window_merged.size == 0:
            continue
        tmp = result[result["chr"] == chr].reset_index(drop=True)
        for i, region in enumerate(window_merged):
            merged_start, merged_end = region
            tmp_sorted = tmp.sort_values(by=["center"], ascending=True).reset_index(
                drop=True
            )
            tmp_res = tmp_sorted.loc[int(merged_start) : int(merged_end)].copy()
            if not tmp_res.empty:
                tmp_res["region"] = i + 1
                tmp_res = tmp_res.sort_values(by=["padj_norm"], ascending=False)
                region_gene = pd.concat([region_gene, tmp_res], axis=0)
    region_gene = region_gene.reset_index(drop=True)
    return region_gene


def sliding_window(
    result: DataFrame,
    window: int,
    step: int,
    threshold: np.float64,
) -> Tuple[List[Tuple[np.str_, MatrixFloat64, MatrixFloat64]], DataFrame]:
    """
    Convert the training results to dataframe and calculate the sliding window and merge significant regions

    Parameters
    ----------
    result : DataFrame
        The integrated training results containing the correlation, p-value, and gene information
    window_size : int
        The size of the window
    step : int
        The step size for the sliding window
    threshold : np.float64
        The threshold to filter the mean values

    Returns
    -------
    sliding_window_result : List[np.str, MatrixFloat64, MatrixFloat64]
        The sliding window results is a list of tuples containing the chromosome, the mean values for each window, and the merged windows
    significant_genes : DataFrame
        The significant genes in the green region of the graph
    """

    chr = result["chr"].unique()
    threshold_norm = -np.log10(threshold)
    sw_res = []
    for c in chr:
        window_mean = cal_sliding_window(result, c, window, step)
        window_merged = merge_window(window_mean, threshold_norm)
        sw_res.append((c, window_mean, window_merged))
    sig_genes = significance(sw_res, result, threshold_norm)
    return sw_res, sig_genes


def cal_sliding_window_quantile(
    met: DataFrame,
    chrom: str,
    center_window_kb: int = None,
    center_step_genes: int = None,
    q: float = 0.9,
) -> MatrixFloat64:
    """
    Sliding window based on padj_norm quantile (e.g. 90% quantile) instead of mean.

    This is used by the new quantile-based method that does not require
    an explicit p-value threshold.
    """
    window_score = []
    met_chr = (
        met[met["chr"] == chrom]
        .sort_values(by=["center"], ascending=True)
        .reset_index(drop=True)
    )
    if len(met_chr) == 0:
        return np.array([])

    window_radius_bp = center_window_kb * 1000
    step_genes = center_step_genes

    gene_num = len(met_chr)
    gene_centers = met_chr["center"].values

    i = 0
    while i < gene_num:
        center_gene_center = gene_centers[i]
        window_left = center_gene_center - window_radius_bp
        window_right = center_gene_center + window_radius_bp

        mask = (gene_centers >= window_left) & (gene_centers <= window_right)
        window_genes_indices = np.where(mask)[0].tolist()
        if len(window_genes_indices) == 0:
            i += step_genes
            if i >= gene_num:
                break
            continue

        window_genes = met_chr.iloc[window_genes_indices]
        padj_norm = window_genes["padj_norm"].to_numpy()
        finite_vals = padj_norm[np.isfinite(padj_norm)]
        if finite_vals.size == 0:
            i += step_genes
            if i >= gene_num:
                break
            continue
        score_val = float(np.quantile(finite_vals, q))

        start_idx = window_genes_indices[0]
        end_idx = window_genes_indices[-1]
        window_score.append(np.array([start_idx, end_idx, score_val]))

        i += step_genes
        if i >= gene_num:
            break

    return np.array(window_score) if window_score else np.array([])


def merge_high_windows_by_window_index(
    window_mean: MatrixFloat64, threshold: float
) -> MatrixFloat64 | None:
    if window_mean is None or getattr(window_mean, "size", 0) == 0:
        return None

    mask_high = window_mean[:, 2] >= threshold
    high_indices = np.where(mask_high)[0]
    if high_indices.size == 0:
        return None

    seg_start = int(high_indices[0])
    seg_prev = int(high_indices[0])
    segments: list[tuple[int, int]] = []
    for idx in high_indices[1:]:
        idx = int(idx)
        if idx == seg_prev + 1:
            seg_prev = idx
        else:
            segments.append((seg_start, seg_prev))
            seg_start = seg_prev = idx
    segments.append((seg_start, seg_prev))

    merged: list[np.ndarray] = []
    for win_lo, win_hi in segments:
        s_idx = int(window_mean[win_lo : win_hi + 1, 0].min())
        e_idx = int(window_mean[win_lo : win_hi + 1, 1].max())
        merged.append(np.array([s_idx, e_idx], dtype=float))
    return np.array(merged) if merged else None


def build_qtl_regions_from_high_windows(
    sliding_window_result: List[Tuple[str, MatrixFloat64, MatrixFloat64 | None]],
    result: DataFrame,
    window_threshold: float,
    use_center: bool = True,
) -> DataFrame:
    rows: list[dict[str, float | int | str]] = []
    if window_threshold is None:
        return pd.DataFrame()
    for chr_name, window_mean, _ in sliding_window_result:
        if window_mean is None or getattr(window_mean, "size", 0) == 0:
            continue
        mask_high = window_mean[:, 2] >= window_threshold
        high_indices = np.where(mask_high)[0]
        if high_indices.size == 0:
            continue
        df_chr = result[result["chr"].astype(str) == str(chr_name)]
        if df_chr.empty:
            continue
        if not {"start_min", "end_max"}.issubset(df_chr.columns):
            continue
        if use_center and "center" in df_chr.columns:
            df_chr = df_chr.sort_values(by=["center"], ascending=True).reset_index(
                drop=True
            )
            region_idx = 1
            seg_start = int(high_indices[0])
            seg_prev = int(high_indices[0])
            segments: list[tuple[int, int]] = []
            for idx in high_indices[1:]:
                idx = int(idx)
                if idx == seg_prev + 1:
                    seg_prev = idx
                else:
                    segments.append((seg_start, seg_prev))
                    seg_start = seg_prev = idx
            segments.append((seg_start, seg_prev))
            for win_lo, win_hi in segments:
                s_idx = int(window_mean[win_lo : win_hi + 1, 0].min())
                e_idx = int(window_mean[win_lo : win_hi + 1, 1].max())
                if s_idx < 0 or e_idx >= len(df_chr) or s_idx > e_idx:
                    continue
                genes = df_chr.iloc[s_idx : e_idx + 1].copy()
                if genes.empty:
                    continue
                qtl_start = float(genes["start_min"].min())
                qtl_end = float(genes["end_max"].max())
                gene_count = int(genes.shape[0])
                best_padj = (
                    float(genes["padj"].min()) if "padj" in genes.columns else np.nan
                )
                best_padj_norm = (
                    float(genes["padj_norm"].max())
                    if "padj_norm" in genes.columns
                    else np.nan
                )
                rows.append(
                    {
                        "chr": str(chr_name),
                        "region": region_idx,
                        "qtl_start": qtl_start,
                        "qtl_end": qtl_end,
                        "gene_count": gene_count,
                        "best_padj": best_padj,
                        "best_padj_norm": best_padj_norm,
                    }
                )
                region_idx += 1
    qtl_regions = pd.DataFrame(rows)
    if qtl_regions.empty:
        return qtl_regions
    qtl_regions = qtl_regions.sort_values(
        by=["chr", "region"], ascending=True
    ).reset_index(drop=True)
    qtl_regions["qtl_center"] = (qtl_regions["qtl_start"] + qtl_regions["qtl_end"]) / 2
    for col in ("qtl_start", "qtl_end", "qtl_center"):
        if col in qtl_regions.columns:
            qtl_regions[col] = qtl_regions[col].round().astype(int)
    return qtl_regions


def sliding_window_newmethod(
    result: DataFrame,
    center_window_kb: int = None,
    center_step_genes: int = None,
    q: float = 0.9,
    top_prop: float = 0.10,
) -> Tuple[List[Tuple[np.str_, MatrixFloat64, MatrixFloat64]], DataFrame, float]:
    """
    New quantile-based method:
    - Gene score: padj_norm = -log10(padj)
    - Window score: q-quantile (e.g. 90%) of padj_norm in the window
    - QTL regions: top `top_prop` (e.g. 10%) windows genome-wide by window score

    Returns sliding window results, significant genes and the window-score threshold
    (in padj_norm scale).
    """
    chr_list = result["chr"].unique()
    sw_res = []
    all_scores = []

    for c in chr_list:
        window_score = cal_sliding_window_quantile(
            result,
            c,
            center_window_kb,
            center_step_genes,
            q,
        )
        if window_score.size > 0:
            all_scores.extend(window_score[:, 2].tolist())
        sw_res.append((c, window_score, None))

    if not all_scores:
        return sw_res, result.iloc[0:0].copy(), 0.0

    all_scores_arr = np.array(all_scores, dtype=float)
    finite_scores = all_scores_arr[np.isfinite(all_scores_arr)]
    if finite_scores.size == 0:
        return sw_res, result.iloc[0:0].copy(), 0.0
    window_threshold = float(np.quantile(finite_scores, 1.0 - top_prop))
    sw_res_new = []
    for c, window_score, _ in sw_res:
        if window_score.size == 0:
            sw_res_new.append((c, window_score, None))
            continue
        window_merged = merge_high_windows_by_window_index(
            window_score, window_threshold
        )
        sw_res_new.append((c, window_score, window_merged))
    sig_genes = significance(sw_res_new, result)

    return sw_res_new, sig_genes, window_threshold



import subprocess
import pandas as pd
import numpy as np
import os
import re
from pandas import DataFrame
import importlib
import inspect
from sklearn.base import RegressorMixin


def get_class_from_path(class_path_string: str) -> RegressorMixin:
    """
    Given a string representing a class path, import the class and return it

    Parameters
    ----------
    class_path_string : str
        A string representing the class path, e.g. "module.submodule.ClassName"

    Returns
    -------
    RegressorMixin
        The imported class object, which should be a subclass of RegressorMixin
    """
    if not class_path_string:
        raise ValueError("class_path_string must not be empty")

    if "." in class_path_string:
        module_path, class_name = class_path_string.rsplit(".", 1)
    else:
        model_name = {
            "DecisionTreeRegressor": "sklearn.tree.DecisionTreeRegressor",
            "RandomForestRegressor": "sklearn.ensemble.RandomForestRegressor",
            "SVR": "sklearn.svm.SVR",
        }
        if class_path_string in model_name:
            module_path, class_name = model_name[class_path_string].rsplit(".", 1)
        else:
            raise ValueError(
                f"Invalid class path string '{class_path_string}'. "
                "It should be in the format 'module.submodule.ClassName'"
            )

    try:
        imported_module = importlib.import_module(module_path)
    except ImportError as e:
        raise ImportError(
            f"Failed to import module '{module_path}' from class path '{class_path_string}': {e}"
        )

    try:
        class_object = getattr(imported_module, class_name)
    except AttributeError:
        raise ImportError(
            f"Attribute or class named '{class_name}' not found in module '{module_path}' "
            f"(from class path '{class_path_string}')."
        )

    # Check if the class is a subclass of RegressorMixin
    if not inspect.isclass(class_object) or not issubclass(
        class_object, RegressorMixin
    ):
        raise TypeError(
            f"The class '{class_name}' in module '{module_path}' is not a subclass of RegressorMixin."
        )

    return class_object


def run_plink(cmd: str) -> str:
    """
    Run a plink command and return the output.
    """
    try:
        result = subprocess.run(
            cmd,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        return result.stdout
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"{e.stderr}")
    except Exception as e:
        raise RuntimeError(f"An error occurred while running plink: {e}")


def gtf_to_range(gtf_file: str, region: str = "CDS") -> DataFrame:
    """
    Convert GTF file to plink gene range format
    """
    if not os.path.exists(gtf_file):
        raise FileNotFoundError(f"The file {gtf_file} does not exist")
    gtf = pd.read_csv(gtf_file, sep=r"\s+", header=None).drop(columns=[1, 5, 6, 7])
    gtf.columns = ["chr", "region", "start", "end", "note"]
    gtf = gtf[gtf["region"] == region]
    if gtf.empty:
        return None
    gtf["transcript_id"] = gtf["note"].str.extract(r'transcript_id\s+"([^"]+)"')
    gtf["gene_id"] = gtf["note"].str.extract(r'gene_id\s+"([^"]+)"')
    gtf = gtf[["chr", "start", "end", "transcript_id", "gene_id"]]
    return gtf


def gff3_to_range(gff_file: str, region: str = "CDS") -> DataFrame:
    """
    Convert GFF3 file to plink gene range format
    """
    gff = pd.read_csv(
        gff_file,
        sep="\t",
        header=None,
        comment="#",
        usecols=range(9),
        names=[
            "seqid",
            "source",
            "type",
            "start",
            "end",
            "score",
            "strand",
            "phase",
            "attributes",
        ],
    )
    gff_filtered = gff[gff["type"] == region].copy()
    if gff_filtered.empty:
        return None
    gff_filtered = gff_filtered[["seqid", "start", "end", "attributes"]].copy()
    gff_filtered.columns = ["chr", "start", "end", "attributes"]
    gff_filtered["id"] = gff_filtered["attributes"].str.extract(
        r"[Ii][Dd]=([^;]+)", expand=False
    )
    gff_filtered["parent_temp"] = gff_filtered["attributes"].str.extract(
        r"[Pp]arent=([^;]+)", expand=False
    )
    gff_filtered["parent"] = gff_filtered["parent_temp"].fillna(gff_filtered["id"])
    gff_filtered = gff_filtered[["chr", "start", "end", "id", "parent"]]
    return gff_filtered



import numpy as np
from typing import List, Tuple
from .plink import Plink
from ..nda_typing import VectorBool, VectorInt8, MatrixInt8, TensorFloat64

__all__ = ["SNP", "Plink"]


_ml_base = {"AA": 1, "GG": 2, "CC": 3, "TT": 4}


class SNP(Plink):
    """
    SNP data class for handling binary plink data
    """

    def __init__(self, snp_file: str):
        self.plink_prefix: str = snp_file
        super().__init__(self.plink_prefix)

    def get(self, marker: str) -> Tuple[int, VectorInt8]:
        """
        Returns the snp data for a given marker

        Parameters
        ----------
        marker : str
            the marker name

        Returns
        -------
        Tuple[int, VectorInt8]
            the index of the marker and the genotype
        """
        idx: np.int64 = self.marker2idx(marker)
        geno: VectorInt8 = self[idx][1]
        return int(idx), geno

    def _init_encoding_map(self) -> None:
        """
        Initialize the ml encoding map of the snp data
        """
        self._allele_ml_encoding: MatrixInt8 = np.zeros(
            self._allele_encoding.shape, dtype=np.int8
        )
        for k, v in _ml_base.items():
            self._allele_ml_encoding[np.where(self._allele_encoding == k)] = v

    @property
    def samples(self):
        """
        Return the samples in the snp data
        """
        return self._fam["iid"].to_list()

    @property
    def chrom(self):
        """
        Return the chromosome
        """
        return self._bim["chrom"].unique().astype(np.str_)

    def encode(
        self,
        snps: List[Tuple[int, VectorInt8]],
        onehot: bool = False,
        filter: VectorBool = None,
    ) -> MatrixInt8 | TensorFloat64:
        """
        Encode the snp data to ml encoding

        Parameters
        ----------
        snps : List[Tuple[int, VectorInt8]]
            the snp data list to encode, each element is a tuple of (index, genotype)
        onehot : bool
            whether to convert to onehot encoding
        filter : VectorBool
            the index ndarray to filter the snp data

        Returns
        -------
        MatrixInt8 or TensorFloat64
            the encoded snp data
            - use onehot: TensorFloat64, shape (n_samples, n_snps, 4)
            - not use onehot: MatrixInt8, shape (n_samples, n_snps)
        """
        if not hasattr(self, "_allele_ml_encoding"):
            self._init_encoding_map()
        result: MatrixInt8 = np.array(
            [self._allele_ml_encoding[i][snp] for i, snp in snps]
        )
        if filter is not None:
            result: MatrixInt8 = result[:, filter]
        if onehot:
            result: TensorFloat64 = SNP.convert_onehot(result)
            result: TensorFloat64 = np.transpose(result, (1, 0, 2))
        else:
            result: MatrixInt8 = result.T
        return result

    @staticmethod
    def convert_onehot(X: MatrixInt8) -> TensorFloat64:
        result: TensorFloat64 = np.zeros((X.shape[0], X.shape[1], 4), dtype=float)
        # A G C T
        for i in range(0, 4):
            rows, cols = np.where(X == i + 1)
            result[rows, cols, i] = 1
        return result



from .snp import SNP
from .trait import Trait
from .gene import Gene
import pandas as pd
import numpy as np
from pandas import DataFrame
from typing import List, Tuple
from ..nda_typing import VectorInt8, VectorBool


class Dataset:
    """
    The Dataset class is used to load and manage the SNP, trait, and gene data
    """

    def __init__(self, snp_file=None, gene_file=None, trait_file=None):

        self._init_gene(gene_file)
        self._init_snp(snp_file)
        self.check_chrom()
        if trait_file is not None:
            self._init_trait(trait_file)
            self.trait.filter_df(self.snp._fam)

    def _init_trait(self, trait_file):
        if trait_file is None:
            raise ValueError("Trait file is required")
        self.trait = Trait(trait_file)

    def _init_gene(self, gene_file):
        if gene_file is None:
            raise ValueError("Gene file is required")
        self.gene = Gene(gene_file)

    def _init_snp(self, snp_file):
        if snp_file is None:
            raise ValueError("SNP file is required")
        self.snp = SNP(snp_file)

    def check_chrom(self) -> None:
        """
        Check if the chromosome names match
        """
        chrom_snp = set(self.snp.chrom)
        chrom_gene = set(self.gene.chrom)
        if not chrom_gene.issubset(chrom_snp):
            err_chr = []
            for chr_g in chrom_gene:
                if chr_g not in chrom_snp:
                    err_chr.append(chr_g)
            raise ValueError(
                f"The chromosomes in gene range file: \n{" ".join(err_chr)} \nare not in the SNP file, Please check the chromosome names in both files to ensure they match (or at least that all chromosomes in the gene range file are present in the SNP file)"
            )

    def get(self, gene: str) -> List[Tuple[int, VectorInt8]]:
        """
        Return the snps for a given gene

        Parameters
        ----------
        gene : str
            The gene name

        Returns
        -------
        List[Tuple[int, VectorInt8]]
            A list of tuples, each tuple contains the snp index and the binary snp data
        """
        gene_position: DataFrame = self.gene.df[self.gene.df["gene"] == gene].filter(
            ["chr", "start", "end"]
        )
        min_start: int = gene_position["start"].min()
        max_end: int = gene_position["end"].max()
        mask: VectorBool = (
            (self.snp._bim["chrom"] == gene_position["chr"].values[0])
            & (self.snp._bim["pos"] >= min_start)
            & (self.snp._bim["pos"] <= max_end)
        )
        merged: DataFrame = pd.merge(
            self.snp._bim[["chrom", "pos", "i"]][mask],
            gene_position,
            left_on="chrom",
            right_on="chr",
        )
        if merged.empty:
            raise ValueError(f"The gene {gene} does not exist in the snp data")

        mask: VectorBool = (merged["pos"] >= merged["start"]) & (
            merged["pos"] <= merged["end"]
        )
        markers_idx: List[int] = merged[mask]["i"].to_list()
        if not markers_idx:
            return np.array([])

        return [(i, self.snp._seek_and_read(i)) for i in markers_idx]

    def get_hap(self, gene: str) -> DataFrame:
        """
        Return the haplotype for a given gene
        """
        gene_snps = self.get(gene)
        gene_snps_double_base = np.array([self.snp.base(*s) for s in gene_snps])
        double_to_single = {
            "AA": "A",
            "TT": "T",
            "CC": "C",
            "GG": "G",
            "AG": "R",
            "TC": "Y",
            "AC": "M",
            "GT": "K",
            "GC": "S",
            "AT": "W",
            "00": "-",
        }
        hap = np.full(
            gene_snps_double_base.shape, "N", dtype=gene_snps_double_base.dtype
        )
        for double, single in double_to_single.items():
            hap[gene_snps_double_base == double] = single
        hap = (
            pd.Series(["".join(h) for h in hap.T])
            .value_counts()
            .to_frame()
            .reset_index()
        )
        hap.columns = ["seq", "count"]
        hap.index.name = "rank"
        hap.index = hap.index + 1
        hap["hap"] = [f"Hap{idx}" for idx in hap.index.values]
        hap = hap.filter(["hap", "seq", "count"])
        return hap



from typing import Tuple, Annotated
import numpy as np
from numpy.typing import NDArray


VectorInt8 = Annotated[NDArray[np.int8], Tuple[int]]
"""1D array of 8-bit signed integers."""

VectorFloat64 = Annotated[NDArray[np.float64], Tuple[int]]
"""1D array of 64-bit floating-point numbers (double precision)."""

VectorStr = Annotated[NDArray[np.str_], Tuple[int]]
"""1D array of NumPy Unicode strings."""

VectorBool = Annotated[NDArray[np.bool_], Tuple[int]]
"""1D array of NumPy booleans."""


MatrixInt8 = Annotated[NDArray[np.int8], Tuple[int, int]]
"""2D array (matrix) of 8-bit signed integers."""

MatrixFloat64 = Annotated[NDArray[np.float64], Tuple[int, int]]
"""2D array (matrix) of 64-bit floating-point numbers (double precision)."""


TensorFloat64 = Annotated[NDArray[np.float64], Tuple[int, int, int]]
"""3D array (tensor) of 64-bit floating-point numbers (double precision)."""



import os
import numpy as np
import pandas as pd
from typing import List, Tuple
from pandas import DataFrame
from ..nda_typing import VectorStr, VectorBool

__all__ = ["Gene"]


class Gene:
    """
    Gene class for handling gene position information
    """

    def __init__(self, file: str):
        if not os.path.exists(file):
            raise FileNotFoundError(f"The file {file} does not exist.")
        self.df: DataFrame = pd.read_csv(file, sep=r"\s+", header=None)
        self.df.columns = ["chr", "start", "end", "transcript", "gene"]
        self.name: VectorStr = self.df["gene"].unique().astype(np.str_)
        self.df = self.df.astype(
            {
                "chr": np.str_,
                "start": np.int64,
                "end": np.int64,
                "transcript": np.str_,
                "gene": np.str_,
            }
        )

    def __repr__(self):
        """
        String representation of the Gene class
        """
        return f"Gene: total {len(self.name)} genes"

    def __len__(self):
        """
        Length of the gene list
        """
        return len(self.name)

    def __getitem__(self, key) -> np.str_:
        """
        Get the gene name by index or name
        """
        return self.name[key]

    def __iter__(self):
        """
        Iterate over the gene names
        """
        return iter(self.name)

    def get(self, gene: str) -> DataFrame:
        """
        Retrieve the location information for a specific gene
        """
        mask: VectorBool = self.df["gene"] == gene
        return self.df[mask]

    def filter(self, genes: List[str]) -> None:
        """
        Filter the gene dataframe to only include the specified genes
        """
        self.df: DataFrame = self.df[self.df["gene"].isin(genes)]
        self.name: VectorStr = self.df["gene"].unique().astype(np.str_)

    def filter_by_chr(self, chr: List[str]) -> None:
        """
        Filter the gene dataframe to only include genes on a specific chromosome
        """
        self.df: DataFrame = self.df[self.df["chr"].isin(chr)].reset_index(drop=True)
        self.name: VectorStr = self.df["gene"].unique().astype(np.str_)

    def chunks(self, p: int) -> List[VectorBool]:
        """
        Split the gene list into chunks of size p

        Parameters
        ----------
        p : int
            the number of chunks to split the gene list into

        Returns
        -------
        List[VectorBool]:
            a list of boolean arrays which can use to index the gene name ndarray
        """
        num: int = int(np.ceil(len(self) / p))
        chunks: List[VectorBool] = [
            pd.Series(self.name).isin(self.name[i : i + num]).to_numpy()
            for i in range(0, len(self), num)
        ]
        return chunks

    @property
    def chrom(self):
        """
        Return the chromosome
        """
        return self.df["chr"].unique().astype(np.str_)



import click
import os
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

console = Console()


def _print_params_panel(rows):
    table = Table(show_header=True, header_style="bold", box=box.SIMPLE)
    table.add_column("Parameter", style="bold")
    table.add_column("Value", style="cyan")
    for key, value in rows:
        table.add_row(key, value)
    panel = Panel.fit(table, title="mlQTL Analysis Parameters", border_style="green")
    console.print("\n", panel, "\n", sep="")


def _info(message):
    console.print(f"[bold]INFO:[/bold] {message}")


def _warn(message):
    console.print(f"[bold yellow]WARNING:[/bold yellow] {message}")


def _error(message):
    console.print(f"[bold red]ERROR:[/bold red] {message}")


def _success(message):
    console.print(f"[bold green]SUCCESS:[/bold green] {message}")


# from .data import Dataset
# from .train import train_with_progressbar, feature_importance
# from .datautils import (
#     sliding_window_newmethod,
#     proc_train_res,
#     build_qtl_regions_from_high_windows,
# )
# from .plot import plot_graph, plot_feature_importance
# from .utils import get_class_from_path, gff3_to_range, gtf_to_range


@click.group()
def main():
    """mlQTL: Machine Learning for QTL Analysis"""
    pass


@main.command()
@click.option(
    "-g",
    "--geno",
    type=str,
    required=True,
    help="Path to genotype file (plink binary format)",
)
@click.option(
    "-p",
    "--pheno",
    type=click.Path(exists=True),
    required=True,
    help="Path to phenotype file",
)
@click.option(
    "-r",
    "--range",
    type=click.Path(exists=True),
    required=True,
    help="Path to plink gene range file",
)
@click.option(
    "-o",
    "--out",
    type=click.Path(),
    required=True,
    help="Path to output directory",
)
@click.option(
    "-j",
    "--jobs",
    type=int,
    default=1,
    help="Number of processes to use",
    show_default=True,
)
@click.option(
    "-m",
    "--model",
    type=str,
    default="DecisionTreeRegressor,RandomForestRegressor,SVR",
    help="Model to use",
    show_default=True,
)
@click.option("-c", "--chrom", type=str, default=None, help="Chromosome to analyze")
@click.option("--trait", type=str, default=None, help="Trait to analyze")
@click.option(
    "--onehot",
    is_flag=True,
    default=False,
    help="Use one-hot encoding for categorical features",
)
@click.option(
    "--padj",
    type=bool,
    default=True,
    show_default=True,
    help="Use adjusted p-value for significance threshold",
)
@click.option(
    "--center-window-kb",
    type=int,
    default=400,
    show_default=True,
    help="Window radius in kilobases (kb) for symmetric neighborhood (e.g., 400 for ±400kb)",
)
@click.option(
    "--center-step-genes",
    type=int,
    default=10,
    show_default=True,
    help="Step size in number of genes for center-based window",
)
@click.option(
    "--q",
    type=float,
    default=0.9,
    show_default=True,
    help="Quantile used as window score for --newmethod (e.g. 0.9 = 90% quantile)",
)
@click.option(
    "--top-prop",
    type=float,
    default=0.10,
    show_default=True,
    help="Top proportion of windows genome-wide selected as QTL for --newmethod (e.g. 0.10 = top 10%)",
)
def run(
    geno,
    pheno,
    range,
    out,
    jobs,
    model,
    chrom,
    trait,
    onehot,
    padj,
    center_window_kb,
    center_step_genes,
    q,
    top_prop,
):
    """Run mlQTL analysis"""
    from .data import Dataset
    from .train import train_with_progressbar
    from .datautils import (
        sliding_window_newmethod,
        proc_train_res,
        build_qtl_regions_from_high_windows,
    )
    from .plot import plot_graph
    from .utils import get_class_from_path

    _print_params_panel(
        [
            ("Genotype file", str(geno)),
            ("Phenotype file", str(pheno)),
            ("Gene range file", str(range)),
            ("Output directory", str(out)),
            ("Number of processes", str(jobs)),
            ("Model(s)", str(model)),
            (
                "Sliding window method",
                f"Symmetric neighborhood (±{center_window_kb} kb around each gene)",
            ),
            ("Sliding window step", f"{center_step_genes} genes"),
            ("Window score quantile (q)", str(q)),
            ("QTL definition (top windows genome-wide)", str(top_prop)),
            ("Chromosome", chrom if chrom else "all chromosomes"),
            ("Trait", trait if trait else "all traits"),
            ("One-hot encoding", "enabled" if onehot else "disabled"),
            ("Use adjusted p-value", "enabled" if padj else "disabled"),
        ]
    )

    try:
        dataset = Dataset(geno, range, pheno)
    except Exception as e:
        _error(str(e))
        return
    max_workers = jobs if jobs > 0 else 1
    analysis_trait = dataset.trait.name
    model = model.split(",")

    try:
        models = [get_class_from_path(m) for m in model]
    except ImportError as e:
        _error(str(e))
        return

    # check option values
    if trait:
        input_trait = set(trait.split(","))
        all_trait = set(dataset.trait.name)
        if not input_trait <= all_trait:
            _error(f"Trait {trait} not found in dataset")
            return
        analysis_trait = trait.split(",")

    if chrom:
        input_chrom = set(chrom.split(","))
        all_chrom = set(dataset.gene.df["chr"].unique())
        if not input_chrom <= all_chrom:
            _error(f"Chromosome {chrom} not found in dataset")
            return
        dataset.gene.filter_by_chr(chrom.split(","))

    # create output directory if not exists
    output_dir = os.path.join(out)
    try:
        os.makedirs(output_dir)
    except FileExistsError:
        _warn(
            f"Output directory {output_dir} already exists. Existing files may be overwritten."
        )
    except OSError as e:
        _error(f"Error creating output directory {output_dir}: {e}")
        return

    # Start the analysis
    _info("Starting analysis")
    for trait in analysis_trait:
        console.print(f"[bold]Trait:[/bold] {trait}")
        _info("Training model")
        train_res = train_with_progressbar(
            trait,
            models,
            dataset,
            max_workers,
            onehot,
        )
        _info("Processing training result")
        try:
            train_res_processed = proc_train_res(train_res, models, dataset, padj)
            sw_res, sig_genes, window_threshold = sliding_window_newmethod(
                train_res_processed,
                center_window_kb,
                center_step_genes,
                q=q,
                top_prop=top_prop,
            )

        except Exception as e:
            _error(str(e))
            return

        if sig_genes.empty:
            _warn("No significant genes found for this trait")

        trait_dir = os.path.join(out, f"{trait}")
        os.mkdir(trait_dir) if not os.path.exists(trait_dir) else None
        # plot and save
        plot_path = os.path.join(trait_dir, "sliding_window")
        plot_graph(sw_res, 10 ** (-window_threshold), plot_path, save=True)
        _info(f"Result graph saved to {plot_path}.png")
        # save the sliding window result
        df_path = os.path.join(trait_dir, "candidate_genes.tsv")
        sig_genes.to_csv(
            df_path,
            sep="\t",
            header=True,
            index=False,
        )
        _info(f"Candidate genes table saved to {df_path}")
        qtl_regions = build_qtl_regions_from_high_windows(
            sw_res,
            train_res_processed,
            window_threshold,
        )
        if not qtl_regions.empty:
            qtl_path = os.path.join(trait_dir, "qtl_regions.tsv")
            qtl_regions.to_csv(qtl_path, sep="\t", header=True, index=False)
            _info(f"QTL regions table saved to {qtl_path}")

        # save the original training result
        # import pickle
        # pkl_path = os.path.join(trait_dir, "train_res.pkl")
        # with open(pkl_path, "wb") as f:
        #     pickle.dump(train_res, f)
        # _info(f"Original training result saved to {pkl_path}")
        # save the training result as dataframe
        train_res_processed.to_csv(
            os.path.join(trait_dir, "train_res.tsv"),
            sep="\t",
            index=False,
            header=True,
        )
        _info(
            f"Training result table saved to {os.path.join(trait_dir, 'train_res.tsv')}"
        )
    _success("Analysis completed")


@main.command()
@click.option(
    "-g",
    "--geno",
    type=str,
    required=True,
    help="Path to genotype file (plink binary format)",
)
@click.option(
    "-p",
    "--pheno",
    type=click.Path(exists=True),
    required=True,
    help="Path to phenotype file",
)
@click.option(
    "-r",
    "--range",
    type=click.Path(exists=True),
    required=True,
    help="Path to plink gene range file",
)
@click.option("-o", "--out", type=click.Path(), required=True, help="Output directory")
@click.option("--gene", type=str, required=True, help="Gene name ( only one gene )")
@click.option(
    "-m",
    "--model",
    type=str,
    default="DecisionTreeRegressor,RandomForestRegressor,SVR",
    help="Model to use",
    show_default=True,
)
@click.option("--trait", type=str, required=True, help="Trait name ( only one trait )")
@click.option(
    "--onehot",
    is_flag=True,
    default=False,
    help="Use one-hot encoding for categorical features",
)
def importance(geno, pheno, range, gene, model, trait, out, onehot):
    """Calculate feature importance and plot bar chart"""
    from .data import Dataset
    from .train import feature_importance
    from .plot import plot_feature_importance
    from .utils import get_class_from_path

    try:
        dataset = Dataset(geno, range, pheno)
    except Exception as e:
        _error(str(e))
        return
    model = model.split(",")
    try:
        models = [get_class_from_path(m) for m in model]
    except ImportError as e:
        _error(str(e))
        return

    # create output directory if not exists
    output_dir = os.path.join(out)
    try:
        os.makedirs(output_dir)
    except FileExistsError:
        _warn(
            f"Output directory {output_dir} already exists. Existing files may be overwritten."
        )
    except OSError as e:
        _error(f"Error creating output directory {output_dir}: {e}")
        return

    if gene not in dataset.gene.name:
        _error(f"Gene {gene} not found in dataset")
        return
    if trait not in dataset.trait.name:
        _error(f"Trait {trait} not found in dataset")
        return
    _info(f"Calculating feature importance for gene {gene} and trait {trait}")
    feature_importance_df = feature_importance(gene, trait, models, dataset, onehot)
    _success("Feature importance calculated successfully")
    gene_dir = os.path.join(output_dir, gene)
    if not os.path.exists(gene_dir):
        os.mkdir(gene_dir)
    # save the feature importance dataframe
    feature_importance_df.to_csv(
        os.path.join(gene_dir, f"{gene}_{trait}_feature_importance.tsv"),
        sep="\t",
        index=True,
    )
    _info("Starting feature importance plot")
    # feature importance plot
    plot_path = os.path.join(gene_dir, f"{gene}_{trait}")
    plot_feature_importance(feature_importance_df, 10, True, plot_path)
    _info(f"Feature importance plot saved to {gene_dir}")


@main.command()
@click.option(
    "-f", "--gff", type=click.Path(exists=True), required=True, help="Path to gff file"
)
@click.option("-r", "--region", type=str, required=True, help="The region to convert")
@click.option("-o", "--out", type=click.Path(), required=True, help="Output directory")
def gff2range(gff, region, out):
    """Convert GFF3 file to plink gene range format"""
    from .utils import gff3_to_range

    try:
        os.makedirs(out)
    except FileExistsError:
        _warn(
            f"Output directory {out} already exists. Existing files may be overwritten."
        )
    df = gff3_to_range(gff, region)
    if df is None:
        _error("The selected area does not exist in the file")
        return
    prefix = os.path.splitext(os.path.basename(gff))[0]
    out_path = os.path.join(out, f"{prefix}_{region}.range")
    df.to_csv(out_path, sep="\t", header=False, index=False)
    _success(f"The range file is saved to {out_path}")


@main.command()
@click.option(
    "-f", "--gtf", type=click.Path(exists=True), required=True, help="Path to gff file"
)
@click.option("-r", "--region", type=str, required=True, help="The region to convert")
@click.option("-o", "--out", type=click.Path(), required=True, help="Output directory")
def gtf2range(gtf, region, out):
    """Convert GTF file to plink gene range format"""

    from .utils import gtf_to_range

    try:
        os.makedirs(out)
    except FileExistsError:
        _warn(
            f"Output directory {out} already exists. Existing files may be overwritten."
        )
    df = gtf_to_range(gtf, region)
    if df is None:
        _error("The selected area does not exist in the file")
        return
    prefix = os.path.splitext(os.path.basename(gtf))[0]
    out_path = os.path.join(out, f"{prefix}_{region}.range")
    df.to_csv(out_path, sep="\t", header=False, index=False)
    _success(f"The range file is saved to {out_path}")


if __name__ == "__main__":
    main()



import os
import numpy as np
import pandas as pd
from typing import Tuple, List
from pandas import DataFrame
from ..nda_typing import VectorFloat64, VectorBool

__all__ = ["Trait"]


class Trait:
    """
    Trait class for handling trait data
    """

    def __init__(self, traits_file: str):
        if not os.path.exists(traits_file):
            raise FileNotFoundError(f"The file {traits_file} does not exist.")
        with open(traits_file, "r") as f:
            header = f.readline().strip().split("\t")
            if len(header) < 2:
                raise ValueError(
                    "The traits file must contain at least one trait column"
                )
            if not header[0] == "sample":
                raise ValueError("The first column of the traits file must be 'sample'")
        self.df: DataFrame = pd.read_csv(traits_file, sep=r"\s+", header=0)
        self.df = self.df.astype(
            {
                "sample": str,
                **{col: np.float64 for col in self.df.columns if col != "sample"},
            }
        )
        self.name: List[str] = self.df.columns[1:].to_list()

    def __repr__(self):
        """
        Returns a string representation of the Trait object
        """
        return f"Trait({len(self.df):,d} samples, {len(self.name):,d} traits)"

    def __len__(self):
        """
        Returns the number of traits in the file
        """
        return len(self.df)

    def __getitem__(self, key) -> str:
        """
        Returns the trait name for a given key
        """
        return self.name[key]

    def __iter__(self):
        """
        Returns an iterator over the trait names
        """
        return iter(self.name)

    def filter_df(self, fam: pd.DataFrame) -> None:
        """
        Filters the trait data to only include samples in the fam file
        """
        self.df = (
            fam.filter(["iid"])
            .merge(self.df, left_on="iid", right_on="sample", how="left", sort=False)
            .drop(columns=["sample"])
        )

    def get(self, name: str) -> Tuple[VectorFloat64, VectorBool]:
        """
        Returns the trait data for a given name

        Parameters
        ----------
        name : str
            The name of the trait

        Returns
        -------
        Tuple[VectorFloat64, VectorBool]
            The trait value data and a boolean mask indicating which samples are not NaN
        """
        if name not in self.name:
            raise ValueError(f"The trait {name} does not exist.")
        data: VectorFloat64 = self.df[name].values
        not_na: VectorBool = ~np.isnan(data)
        data: VectorFloat64 = data[not_na]
        return data, not_na



"""Module that reads binary Plink files."""

# This file is part of pyplink.
#
# The MIT License (MIT)
#
# Copyright (c) 2014 Louis-Philippe Lemieux Perreault
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Louis-Philippe Lemieux Perreault"
__copyright__ = "Copyright 2014 Louis-Philippe Lemieux Perreault"
__license__ = "MIT"

# Modified by: yang1ong
# Date: 2025-04-23
# Description of changes: Removed unused functions to streamline core functionality

import os
import logging
import numpy as np
import pandas as pd
import multiprocessing
from collections import Counter
from itertools import repeat


# The logger
logger = logging.getLogger(__name__)


# The recoding values
_geno_recode = {
    1: -1,  # Unknown genotype
    2: 1,  # Heterozygous genotype
    0: 2,  # Homozygous A1
    3: 0,
}  # Homozygous A2


class Plink(object):
    """Reads and store a set of binary Plink files.

    Args:
        prefix (str): The prefix of the binary Plink files.

    Reads or write binary Plink files (BED, BIM and FAM).
    """

    # The genotypes values
    _geno_values = np.array(
        [[_geno_recode[(i >> j) & 3] for j in range(0, 7, 2)] for i in range(256)],
        dtype=np.int8,
    )

    def __init__(self, prefix: str):
        """Initializes a new Plink instance."""
        # The mode
        self._mode = "r"

        # The bed format
        self._bed_format = "SNP-major"

        # These are the name of the files
        self.bed_filename = f"{prefix}.bed"
        self.bim_filename = f"{prefix}.bim"
        self.fam_filename = f"{prefix}.fam"

        # Checking that all the files exists (otherwise, error...)
        for filename in (self.bed_filename, self.bim_filename, self.fam_filename):
            if not os.path.isfile(filename):
                raise IOError(f"No such file: '{filename}'")

        # The number of markers and samples
        self._nb_markers = None
        self._nb_samples = None

        # Setting BIM and FAM to None
        self._bim = None
        self._fam = None

        # Reading the input files
        self._read_bim()
        self._read_fam()

        # check the BED file
        self._check_bed()

        # index and lock
        self._index = multiprocessing.Value("i", 0)
        self._lock = multiprocessing.Lock()

    def __repr__(self):
        """The representation of the Plink object."""

        return f"Plink({self.nb_samples:,d} samples; " f"{self.nb_markers:,d} markers)"

    def __len__(self):
        """
        Returns the number of snps in the file
        """
        return self.nb_markers

    def __getitem__(self, key: int):
        """The __getitem__ function.
        Args:
            key (int): The index of the marker to read.
        Returns:
            tuple: The marker name as a string and its genotypes as a
            :py:class:`numpy.ndarray`.
        """
        return self.idx2marker(key), self._seek_and_read(key)

    def __iter__(self):
        """The __iter__ function."""
        return self

    def __next__(self):
        """The __next__ function."""
        return self.next()

    def next(self):
        """Returns the next marker.

        Returns:
            tuple: The marker name as a string and its genotypes as a
            :py:class:`numpy.ndarray`.

        """
        with self._lock:
            current_index = self._index.value
            if current_index > self.nb_markers:
                raise StopIteration()
            self._index.value += 1
            return self[current_index]

    def _get_seek_position(self, n: int):
        """Gets the seek position in the file (including special bytes).

        Args:
            n (int): The index of the marker to seek to.

        """
        return 3 + self._nb_bytes * n

    def _seek_and_read(self, n: int) -> np.ndarray:
        """Reads the genotypes of a idx"""
        with open(self.bed_filename, "rb") as bed:
            bed.read(3)
            if n < 0 or n >= self.nb_markers:
                raise ValueError(f"invalid position in BED: {n}")
            bed.seek(self._get_seek_position(n))
            return self._geno_values[
                np.frombuffer(bed.read(self._nb_bytes), dtype=np.uint8)
            ].flatten(order="C")[: self.nb_samples]

    @property
    def nb_markers(self):
        if self._nb_markers is None:
            self._nb_markers = self._bim.shape[0]
        return self._nb_markers

    @property
    def duplicated_markers(self):
        if self._has_duplicated:
            return self._dup_markers
        else:
            return {}

    @property
    def nb_samples(self):
        if self._nb_samples is None:
            self._nb_samples = self._fam.shape[0]
        return self._nb_samples

    def _read_bim(self):
        """Reads the BIM file."""
        # Reading the BIM file and setting the values
        bim = pd.read_csv(
            self.bim_filename,
            sep=r"\s+",
            names=["chrom", "snp", "cm", "pos", "a1", "a2"],
            dtype=dict(chrom=str, snp=str, a1=str, a2=str),
        )

        # replace the empty markers with "chrom-pos"
        bim["snp"] = bim["snp"].replace(".", np.nan).infer_objects(copy=False)
        bim["snp"] = bim["snp"].fillna(bim["chrom"] + "-" + bim["pos"].astype(str))
        bim["snp"] = bim["snp"].astype(str)

        # Saving the index as integer
        bim["i"] = bim.index

        # Checking for duplicated markers
        try:
            bim = bim.set_index("snp", verify_integrity=True)
            self._has_duplicated = False

        except ValueError:
            # Setting this flag to true
            self._has_duplicated = True

            # Finding the duplicated markers
            duplicated = bim.snp.duplicated(keep=False)
            duplicated_markers = bim.loc[duplicated, "snp"]
            duplicated_marker_counts = duplicated_markers.value_counts()

            # The dictionary that will contain information about the duplicated
            # markers
            self._dup_markers = {m: [] for m in duplicated_marker_counts.index}

            # Logging a warning
            logger.warning("Duplicated markers found")
            for marker, count in duplicated_marker_counts.items():
                logger.warning("  - %s: %s times", marker, count)
            logger.warning(
                "Appending ':dupX' to the duplicated markers "
                "according to their location in the BIM file"
            )

            # Renaming the markers
            counter = Counter()
            for i, marker in duplicated_markers.items():
                counter[marker] += 1
                new_name = f"{marker}:dup{counter[marker]}"
                bim.loc[i, "snp"] = new_name

                # Updating the dictionary containing the duplicated markers
                self._dup_markers[marker].append(new_name)

            # Resetting the index
            bim = bim.set_index("snp", verify_integrity=True)

        # Encoding the allele
        #   - The original 0 is the actual 2 (a1/a1)
        #   - The original 2 is the actual 1 (a1/a2)
        #   - The original 3 is the actual 0 (a2/a2)
        #   - The original 1 is the actual -1 (no call)
        allele_encoding = np.array(
            [bim.a2 * 2, bim.a1 + bim.a2, bim.a1 * 2, list(repeat("00", bim.shape[0]))],
            dtype="U2",
        )
        self._allele_encoding = allele_encoding.T

        # Saving the data in the object
        self._bim = bim[["chrom", "pos", "cm", "a1", "a2", "i"]]

    def _read_fam(self):
        """Reads the FAM file."""
        # Reading the FAM file and setting the values
        fam = pd.read_csv(
            self.fam_filename,
            sep=r"\s+",
            names=["fid", "iid", "father", "mother", "gender", "status"],
            dtype=dict(fid=str, iid=str, father=str, mother=str),
        )

        # Saving the data in the object
        self._fam = fam

    def _check_bed(self):
        """check the BED file."""
        # Checking if BIM and BAM files were both read
        if (self._bim is None) or (self._fam is None):
            raise RuntimeError("no BIM or FAM file were read")

        # The number of bytes per marker
        self._nb_bytes = int(np.ceil(self.nb_samples / 4.0))

        # Checking the file is valid by looking at the first 3 bytes and the
        # last entry (correct size)
        with open(self.bed_filename, "rb") as bed_file:
            # Checking that the first two bytes are OK
            if (ord(bed_file.read(1)) != 108) or (ord(bed_file.read(1)) != 27):
                raise ValueError(f"not a valid BED file: {self.bed_filename}")

            # Checking that the format is SNP-major
            if ord(bed_file.read(1)) != 1:
                raise ValueError(
                    f"not in SNP-major format (please recode): " f"{self.bed_filename}"
                )

            # Checking the last entry (for BED corruption)
            seek_index = self._get_seek_position(self._bim.iloc[-1, :].i)
            bed_file.seek(seek_index)
            geno = self._geno_values[
                np.frombuffer(bed_file.read(self._nb_bytes), dtype=np.uint8)
            ].flatten(order="C")[: self.nb_samples]
            if geno.shape[0] != self.nb_samples:
                raise ValueError("invalid number of entries: corrupted BED?")

        # Opening the file for the rest of the operations (reading 3 bytes)
        # self._bed = open(self.bed_filename, "rb")
        # self._bed.read(3)

    def marker2idx(self, marker: str) -> np.int64:
        """Returns the index of a marker"""
        if marker not in self._bim.index:
            raise ValueError(f"{marker}: marker not in BIM")
        return self._bim.loc[marker, "i"]

    def idx2marker(self, idx: int) -> str:
        """Returns the marker name from its index"""
        if idx < 0 or idx >= self.nb_markers:
            raise ValueError(f"{idx}: index out of range")
        return self._bim.index[idx]

    def base(self, key: int | str, snp: np.ndarray) -> np.ndarray:
        """Convert binary to nucleobase"""
        match key:
            case int():
                return self._allele_encoding[key][snp]
            case str():
                return self._allele_encoding[self.marker2idx(key)][snp]
            case _:
                raise TypeError(f"Invalid key type: {type(key)}")



import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame
from typing import List, Tuple
from .nda_typing import MatrixFloat64


def _plot_chr(
    axs,
    chr: str,
    window_mean: MatrixFloat64,
    window_merged: MatrixFloat64 | None,
    threshold: float,
):
    if window_mean is None or getattr(window_mean, "size", 0) == 0:
        axs.set_xlabel("Window Index")
        axs.set_ylabel("-log10(P-value)")
        axs.set_title(f"Chr {chr} (Gene number: 0)")
        return axs

    y = window_mean[:, 2]
    x = np.arange(len(y))

    axs.plot(x, y, color="black")

    mask = y >= threshold
    if mask.any():
        axs.fill_between(x, y, where=mask, color="green", alpha=1)
    axs.axhline(y=threshold, color="#dc7633", linestyle="dashed")
    axs.set_xlabel("Window Index")
    axs.set_ylabel("-log10(P-value)")
    axs.set_title(f"Chr {chr} (Gene number: {int(window_mean[-1][1])})")

    return axs


def plot_graph(
    sliding_window_result: List[Tuple[str, MatrixFloat64, MatrixFloat64]],
    threshold: np.float64,
    filename: str = "result",
    font_size: int = 20,
    save: bool = False,
) -> None:
    """
    Plot the sliding window result
    """
    threshold_norm = -np.log10(threshold)

    plt.rcParams["font.size"] = font_size
    chr_num = len(sliding_window_result)
    _, axs = plt.subplots(chr_num, figsize=(20, 40 / 12 * chr_num))
    try:
        sliding_window_result.sort(key=lambda x: int(x[0]))
    except ValueError:
        sliding_window_result.sort(key=lambda x: x[0])
    if len(sliding_window_result) == 1:
        axs = [axs]
    for i, res in enumerate(sliding_window_result):
        chr, window_mean, window_merged = res
        axs[i] = _plot_chr(axs[i], chr, window_mean, window_merged, threshold_norm)

    plt.tight_layout()
    plt.show()
    if save:
        plt.savefig(f"{filename}.png", dpi=300, bbox_inches="tight")


def plot_feature_importance(
    feature_importance: DataFrame,
    topn: int = 10,
    save: bool = False,
    filename: str = "feature_importance",
) -> None:
    """
    Plot the feature importance of SNPs in different models
    """
    topn = 10
    for row in feature_importance.iterrows():
        model_name = row[0]
        snps = row[1].index.tolist()
        snp_number = {snp: str(i + 1) for i, snp in enumerate(snps)}
        top_n = row[1].sort_values(ascending=False).head(topn)
        y = top_n.values
        x_labels_snp = top_n.index
        x_labels_snp = [f"{snp} ({snp_number[snp]})" for snp in x_labels_snp]
        x_positions = range(len(x_labels_snp))
        plt.style.use("seaborn-v0_8-whitegrid")
        plt.figure(figsize=(8, 6))
        bar_width = 0.4
        bars = plt.bar(
            x_labels_snp,
            y,
            width=bar_width,
            color="steelblue",
            edgecolor="black",
            alpha=1,
        )
        plt.xlabel(f"Top {topn} SNP", fontsize=14)
        plt.xticks(x_positions, x_labels_snp, rotation=45, ha="right", fontsize=10)
        plt.ylabel("Importance Score", fontsize=14)
        plt.title(
            f"SNP Feature Importance in Model: {model_name}",
            fontsize=16,
            fontweight="bold",
            pad=20,
        )

        ax = plt.gca()
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.yaxis.grid(True, linestyle="--", alpha=0.9)
        ax.xaxis.grid(False)
        plt.tight_layout()
        if save:
            plt.savefig(f"{filename}_{model_name}.png", dpi=300, bbox_inches="tight")
        else:
            plt.show()



import numpy as np
import multiprocessing
from typing import List, Tuple
from scipy.stats import pearsonr, NearConstantInputWarning, ConstantInputWarning
from sklearn import preprocessing
from sklearn.model_selection import train_test_split
from sklearn.inspection import permutation_importance
from inspect import signature
from sklearn.base import RegressorMixin
from pandas import DataFrame
import warnings
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

from .data import Dataset
from .nda_typing import (
    MatrixInt8,
    TensorFloat64,
    VectorFloat64,
    MatrixFloat64,
    VectorStr,
)

np.random.seed(42)


class MLMetrics(object):
    def __init__(self):
        self.corr = None
        self.p_value = None

    def update(self, y, y_hat):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NearConstantInputWarning)
            warnings.simplefilter("ignore", ConstantInputWarning)
            with np.errstate(divide="ignore", invalid="ignore"):
                corr, p_value = pearsonr(y, y_hat)
        self.corr = corr
        self.p_value = p_value


def train_batch(
    X: MatrixInt8 | TensorFloat64,
    y: VectorFloat64,
    onehot: bool,
    models: List[RegressorMixin],
    importance: bool = False,
) -> MatrixFloat64:
    """
    Train a batch of models on the given data

    Parameters
    ----------
    X : MatrixInt8 | TensorFloat64
        The encoded SNP data
    y : VectorFloat64
        The trait values
    onehot : bool
        Whether the SNP data is one-hot encoded
    models : List[RegressorMixin]
        The list of models to train
    importance : bool
        Whether to calculate feature importance

    Returns
    -------
    MatrixFloat64
        - importance == False: a matrix of shape (n_models, 2) containing the correlation and p-value for each model
        - importance == True, a matrix of shape (n_models, n_features) containing the feature importance for each model
    """
    if onehot:
        x_shape_ori: Tuple[int, int, int] = X.shape
        X: MatrixFloat64 = X.reshape(y.shape[0], -1)
    else:
        min_max_scaler = preprocessing.MinMaxScaler()
        X = min_max_scaler.fit_transform(X)

    instances = []
    for model in models:
        if not issubclass(model, RegressorMixin):
            raise TypeError(f"Model {model} is not a valid regression model")
        try:
            init_params = signature(model).parameters
            if "random_state" in init_params:
                instances.append(model(random_state=42))
            else:
                instances.append(model())
        except TypeError:
            raise TypeError(f"Model {model} should be a class")

    mets, imp_matrix = [], []
    for model in instances:
        met = MLMetrics()
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2, random_state=42
        )
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)
        met.update(y_test, y_pred)
        mets.append([met.corr, met.p_value])

        if importance:
            imp_matrix.append(
                permutation_importance(
                    model, X_train, y_train, n_repeats=10, random_state=42
                ).importances_mean
            )

    if importance:
        imp_matrix = np.array(imp_matrix)
        if onehot:
            imp_matrix = imp_matrix.reshape(-1, x_shape_ori[1], x_shape_ori[2])
            imp_matrix = np.mean(imp_matrix, axis=2)
        return imp_matrix

    return np.array(mets)


def init_worker(dataset: Dataset) -> None:
    """
    Initialize the worker with the dataset for multiprocessing
    """
    global g_dataset
    g_dataset = dataset


def _task(
    genes: VectorStr, trait: str, onehot: bool, models: RegressorMixin
) -> List[MatrixFloat64 | None]:
    """
    task for each chunk of genes
    """
    result = []
    for gene in genes:
        try:
            snps = g_dataset.get(gene)
            trait_value, not_nan_idx = g_dataset.trait.get(trait)
            X = g_dataset.snp.encode(snps, onehot, filter=not_nan_idx)
            y = trait_value
            m1 = train_batch(X, y, onehot, models, False)
            result.append(m1)
        except ValueError:
            result.append(None)
    return result


def train(
    trait: str,
    models: List[RegressorMixin],
    dataset: Dataset,
    max_workers: int = 8,
    onehot: bool = False,
) -> List[List[MatrixFloat64 | None]]:
    """
    Train models on the given dataset using multiprocessing

    Parameters
    ----------
    trait : str
        The trait to train
    onehot : bool
        Whether the SNP data is one-hot encoded
    models : List[RegressorMixin]
        The list of models to train
    dataset : Dataset
        The dataset to train
    max_workers : int
        The number of workers to use for multiprocessing

    Returns
    -------
    List[List[MatrixFloat64 | None]]
        A list of lists containing the correlation and p-value for each model
        for each gene, shape of MatrixFloat64 is (n_models, 2)
    """
    with multiprocessing.Pool(
        processes=max_workers, initializer=init_worker, initargs=(dataset,)
    ) as pool:
        results = pool.starmap(
            _task,
            [
                (
                    dataset.gene.name[chunk],
                    trait,
                    onehot,
                    models,
                )
                for chunk in dataset.gene.chunks(max_workers)
            ],
        )

    return results


def train_single(
    gene: str,
    trait: str,
    models: List[RegressorMixin],
    dataset: Dataset,
    onehot: bool = False,
) -> MatrixFloat64:
    """
    Train a single gene and trait

    Parameters
    ----------
    gene : str
        The gene to train
    trait : str
        The trait to train
    dataset : Dataset
        The dataset to train
    models : List[RegressorMixin]
        The list of models to train
    onehot : bool
        Whether the SNP data is one-hot encoded

    Returns
    -------
    MatrixFloat64
        a matrix of shape (n_models, 2) containing the correlation and p-value for each model

    """
    snps = dataset.get(gene)
    trait_value, not_nan_idx = dataset.trait.get(trait)
    X = dataset.snp.encode(snps, onehot, filter=not_nan_idx)
    y = trait_value
    res = train_batch(X, y, onehot, models)
    return res


def feature_importance(
    gene: str,
    trait: str,
    models: List[RegressorMixin],
    dataset: Dataset,
    onehot: bool = False,
) -> DataFrame:
    """
    Train a single gene and trait

    Parameters
    ----------
    gene : str
        The gene to train
    trait : str
        The trait to train
    dataset : Dataset
        The dataset to train
    models : List[RegressorMixin]
        The list of models to train
    onehot : bool
        Whether the SNP data is one-hot encoded

    Returns
    -------
    DataFrame
        a pandas DataFrame containing the feature importance for each model
        with the SNP markers as columns and the model names as rows

    """
    snps = dataset.get(gene)
    trait_value, not_nan_idx = dataset.trait.get(trait)
    X = dataset.snp.encode(snps, onehot, filter=not_nan_idx)
    y = trait_value
    importance = train_batch(X, y, onehot, models, True)
    feature = np.array([dataset.snp.idx2marker(i[0]) for i in snps], dtype=np.str_)
    models_name = np.array([model.__name__ for model in models], dtype=np.str_)
    res = DataFrame(
        importance,
        index=models_name,
        columns=feature,
    )
    return res


# The following defines the train function with a progress bar and related helper functions used in the CLI


def _progress_bar_manager(queue: multiprocessing, total_genes: int, trait: str):
    """
    Manages the Rich progress bar by listening to a queue.
    Receives 'None' to terminate.
    """
    columns = [
        TextColumn("[bold]{task.description}[/bold]"),
        BarColumn(),
        TaskProgressColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
    ]
    with Progress(*columns) as progress:
        task_id = progress.add_task(trait, total=total_genes)
        processed_count = 0
        while processed_count < total_genes:
            message = queue.get()
            if message is None:
                break
            if message == "TICK":
                progress.update(task_id, advance=1)
                processed_count += 1
        if processed_count < total_genes:
            progress.update(task_id, completed=processed_count)
        progress.refresh()


def _task_progressbar(
    genes: VectorStr,
    trait: str,
    onehot: bool,
    models: RegressorMixin,
    progress_queue: multiprocessing.Queue,
) -> List[MatrixFloat64 | None]:
    """
    task for each chunk of genes
    """
    result = []
    for gene in genes:
        try:
            snps = g_dataset.get(gene)
            trait_value, not_nan_idx = g_dataset.trait.get(trait)
            X = g_dataset.snp.encode(snps, onehot, filter=not_nan_idx)
            y = trait_value
            m1 = train_batch(X, y, onehot, models, False)
            result.append(m1)
        except ValueError:
            result.append(None)
        finally:
            progress_queue.put("TICK")
    return result


def train_with_progressbar(
    trait: str,
    models: List[RegressorMixin],
    dataset: Dataset,
    max_workers: int = 8,
    onehot: bool = False,
) -> List[List[MatrixFloat64 | None]]:
    """
    Train models on the given dataset using multiprocessing

    Parameters
    ----------
    trait : str
        The trait to train
    onehot : bool
        Whether the SNP data is one-hot encoded
    models : List[RegressorMixin]
        The list of models to train
    dataset : Dataset
        The dataset to train
    max_workers : int
        The number of workers to use for multiprocessing

    Returns
    -------
    List[List[MatrixFloat64 | None]]
        A list of lists containing the correlation and p-value for each model
        for each gene, shape of MatrixFloat64 is (n_models, 2)
    """
    with multiprocessing.Manager() as manager:
        progress_queue = manager.Queue()
        progress_manager_proc = multiprocessing.Process(
            target=_progress_bar_manager,
            args=(progress_queue, len(dataset.gene.name), trait),
        )
        progress_manager_proc.start()

        with multiprocessing.Pool(
            processes=max_workers, initializer=init_worker, initargs=(dataset,)
        ) as pool:
            results = pool.starmap(
                _task_progressbar,
                [
                    (
                        dataset.gene.name[chunk],
                        trait,
                        onehot,
                        models,
                        progress_queue,
                    )
                    for chunk in dataset.gene.chunks(max_workers)
                ],
            )
        progress_queue.put(None)
        progress_manager_proc.join()
    return results
